/*
 * Test IOCTL_PRIVCMD_RESTRICT restricts properly.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/types.h>
#include <xen/xen.h>
#include <xen/sys/privcmd.h>

#include <xen/hvm/dm_op.h>

enum result {
	SUCCESS, /* test successful */
	FAIL,    /* test failed */
	ERROR,   /* error during test preparation */
};

static void done(enum result r, const char *m)
{
	static const char *results[] = {
		[SUCCESS] = "Success",
		[FAIL] = "FAIL",
		[ERROR] = "ERROR",
	};

	printf("%s", results[r]);
	if (m)
		printf(": %s", m);
	if (r != SUCCESS && errno)
		printf(": %s", strerror(errno));
	printf("\n");

	exit(r);
}

static void success(void)
{
	done(SUCCESS, NULL);
}

static void fail(const char *m)
{
	done(FAIL, m);
}

static void error(const char *m)
{
	done(ERROR, m);
}


int main(void)
{
	int fd;
	int ret;
	domid_t restrict_domid;
	privcmd_dm_op_t dmop;
	privcmd_dm_op_buf_t ubuf;
	struct xen_dm_op op;

	fd = open("/dev/xen/privcmd", O_RDWR);
	if (fd < 0)
		error("open /dev/xen/privcmd");

	/*
	 * Use a domid that is unlikely to exist on the system so if
	 * the hypercall goes to the hypervisor it will fail with
	 * something other than EPERM.
	 */
	restrict_domid = 0x7FEF;

	ret = ioctl(fd, IOCTL_PRIVCMD_RESTRICT, &restrict_domid);
	if (ret < 0)
		error("IOCTL_PRIVCMD_RESTRICT");

	ret = mlock(&op, sizeof(op));
	if (ret < 0)
		error("mlock");

	/*
	 * 1. Try a DMOP with the wrong domid. The kernel should
	 *    return EPERM.
	 */

	memset(&op, 0, sizeof(op));
	op.op = XEN_DMOP_track_dirty_vram;
	op.u.track_dirty_vram.first_pfn = 0;
	op.u.track_dirty_vram.nr = 0;
	ubuf.uptr = &op;
	ubuf.size = sizeof(op);
	dmop.dom = 0;
	dmop.num = 1;
	dmop.ubufs = &ubuf;

	ret = ioctl(fd, IOCTL_PRIVCMD_DM_OP, &dmop);
	if (ret >= 0)
		fail("dmop successful\n");
	if (errno != EPERM)
		fail("dmop failed with errno != EPERM");

	/*
	 * 2. Use the restricted domid. This should get to Xen and
	 *    return ESRCH since it is a non-existent domid.
	 */

	dmop.dom = restrict_domid;

	ret = ioctl(fd, IOCTL_PRIVCMD_DM_OP, &dmop);
	if (ret >= 0)
		fail("dmop successful\n");
	if (errno != ESRCH)
		fail("dmop failed with errno != ESRCH");

	success();
	return 0;
}
